(require 'asdf)
(require 'pal)

(defpackage :columns
  (:use :common-lisp)
  (:export #:play))

(in-package :columns)

(defparameter +matrix-width+ 6)
(defparameter +matrix-height+ 14)

(defun init-matrix ()
  (make-array (list +matrix-height+ +matrix-width+) :initial-element nil))

(defmacro do-matrix ((i j) &body body)
  `(loop for ,j below +matrix-height+
         do (loop for ,i below +matrix-width+
                  do (progn ,@body))))

(defclass column ()
  ((x :initarg :x :accessor x)
   (y :initarg :y :accessor y)
   (colors :initarg :colors :accessor colors)))

(defun init-column ()
  (make-instance 'column
                 :x (floor (1- +matrix-width+) 2) :y 2
                 :colors (loop repeat 3
                               collect (random 6))))

(defun rotate (column)
  (let ((colors (colors column)))
    (setf (colors column)
          (append (rest colors) (list (first colors))))))

(defparameter +screen-width+ 800)
(defparameter +screen-height+ 600)

(defparameter +matrix-display-x0+ 200)
(defparameter +matrix-display-y0+ 3)

(defparameter +cell-size+ 40)

(defvar *block-images* (make-array 6))

(defun i->x (i)
  (+ (* i +cell-size+) +matrix-display-x0+))

(defun j->y (j)
  (+ (* j +cell-size+) +matrix-display-y0+))

(defun draw-game (matrix column)
  (pal:clear-screen 0 0 0)
  (draw-rectangle (1- +matrix-display-x0+)
                  (1- +matrix-display-y0+)
                  (+ (* +cell-size+ +matrix-width+) 2)
                  (+ (* +cell-size+ +matrix-height+) 2)
                  255 255 255)
  (do-matrix (i j)
    (let ((cell (aref matrix j i)))
      (when cell
        (draw-block (i->x i) (j->y j) cell))))
  (when column
    (loop for i below 3
          do (draw-block (i->x (x column))
                         (j->y (- (y column) i))
                         (nth i (colors column))))))

(defun draw-rectangle (x y width height r g b)
  (pal:draw-polygon (list (pal:v x y)
                          (pal:v x (+ y height))
                          (pal:v (+ x width) (+ y height))
                          (pal:v (+ x width) y))
                    r g b
                    255
                    :fill nil))

(defun draw-block (x y color)
  (pal:draw-image (aref *block-images* color) (pal:v x y)))

(defun draw-lines (matrix)
  (let ((lines (compute-lines matrix)))
    (loop for line in lines
          do (loop for (p1 p2) on line
                   when (and p1 p2)
                   do (pal:draw-line (pal:v (+ (i->x (car p1)) 20)
                                            (+ (j->y (cdr p1)) 20))
                                     (pal:v (+ (i->x (car p2)) 20)
                                            (+ (j->y (cdr p2)) 20))
                                     255 255 255
                                     255
                                     :size 3)))))

(defun in-matrix-p (x y)
  (and (<= 0 x (1- +matrix-width+))
       (<= 0 y (1- +matrix-height+))))

(defun position-available-p (x y matrix)
  (and (in-matrix-p x y)
       (null (aref matrix y x))))

(defun update-column (column matrix)
  (cond ((or (= (y column) (- +matrix-height+ 1))
             (aref matrix (1+ (y column)) (x column)))
         (loop for i below 3
               do (setf (aref matrix (- (y column) i) (x column))
                        (nth i (colors column)))
               finally (return t)))
        (t
         (incf (y column))
         nil)))

(defun compute-lines (matrix)
  (let ((lines '()))
    (loop for j below +matrix-height+
          do (loop for i below +matrix-width+
                   as color = (aref matrix j i)
                   unless (null color)
                   do (loop for (di . dj)
                            in '((-1 . 0) (0 . -1) (-1 . -1) (1 . -1))
                            as line = (brick-line-from matrix i j di dj)
                            when (>= (length line) 3)
                            do (push line lines))))
    lines))
                         
(defun brick-line-from (matrix i j di dj)
  (loop with color = (aref matrix j i)
        for n from 0
        as i1 = (+ i (* n di))
        as j1 = (+ j (* n dj))
        while (and (in-matrix-p i1 j1)
                   (eql (aref matrix j1 i1) color))
        collect (cons i1 j1)))

(defun remove-lines (matrix)
  (let ((lines (compute-lines matrix)))
    (loop for line in lines
          do (loop for (p1 p2) on line
                   do (setf (aref matrix (cdr p1) (car p1)) nil)))))

(defun fall-bricks (matrix)
  (loop for i below +matrix-width+
        do (loop for j from (1- +matrix-height+) downto 1
                 when (and (null (aref matrix j i))
                           (loop for j0 from j downto 0
                                 when (aref matrix j0 i)
                                 return t
                                 finally (return nil)))
                 do (loop while (null (aref matrix j i))
                          do (loop for j0 from j downto 1
                                   do (setf (aref matrix j0 i)
                                            (aref matrix (1- j0) i)))
                             (setf (aref matrix 0 i) nil)))))

(defun play ()
  (pal:with-pal (:width +screen-width+ :height +screen-height+
                 :fullscreenp nil :fps 50)
    (loop for i below (length *block-images*)
          do (setf (aref *block-images* i)
                   (pal:load-image (format nil "Images/block-~A.png" i))))
    (let ((matrix (init-matrix))
          (column nil)
          (state :stopped)
          (frame-count 0)
          (total-frame-count 0)
          (last-key-frame 0)
          (step-length 40))
      (pal:event-loop ()
        (case state
          (:stopped
           (pal:test-keys
             (:key-space (setf state :playing)
                         (setf column (init-column)))
             (:key-q (return-from pal:event-loop))))
          (:playing
           (when (< last-key-frame (- total-frame-count 4))
             (pal:test-keys
               (:key-left (when (and column
                                     (position-available-p (1- (x column))
                                                           (y column)
                                                           matrix))
                            (decf (x column)))
                          (setf last-key-frame total-frame-count))
               (:key-right (when (and column
                                      (position-available-p (1+ (x column))
                                                            (y column)
                                                            matrix))
                             (incf (x column)))
                           (setf last-key-frame total-frame-count))
               (:key-up (when column
                          (rotate column))
                        (setf last-key-frame total-frame-count))
               (:key-down (when column
                            (setf frame-count step-length)))
               (:key-q (return-from pal:event-loop))))
           (when column
             (incf frame-count)
             (when (> frame-count step-length)
               (when (update-column column matrix)
                 (let ((lines (compute-lines matrix)))
                   (if lines
                       (setf state :line-animation)
                       (setf column (init-column)))))
               (setf frame-count 0)))
           (draw-game matrix column))
          (:line-animation
           (draw-game matrix nil)
           (when (< frame-count 40)
             (draw-lines matrix))
           (incf frame-count)
           (when (> frame-count 50)
             (remove-lines matrix)
             (fall-bricks matrix)
             (when (null (compute-lines matrix))
               (setf state :playing
                     column (init-column)))
             (setf frame-count 0))))
        (incf total-frame-count)))))
